#ifdef WIN32
#pragma warning(disable : 4996)
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "yuv.h"
#include "upsample.h"

#define CLIP3(x, min, max) ((unsigned)(x)<(unsigned)(min)?(unsigned)(min):((unsigned)(x)>(unsigned)(max)?(unsigned)(max):(unsigned)(x)))

CIYuv::CIYuv()
{
  Y = U = V = Ytemp = NULL;
  comp[0] = comp[1] = comp[2] = NULL;
  pBuffer = NULL;
  FuncFilter = NULL;
}


CIYuv::CIYuv(int h, int w, int chroma_format)
{
  height = h;
  width = w;
  sampling = chroma_format;

  Y = U = V = Ytemp = NULL;
  comp[0] = comp[1] = comp[2] = NULL;
  pBuffer = NULL;
  FuncFilter = NULL;

  switch(sampling)
  {
  case 400: // padding chroma components
    heightUV = height/2;
    widthUV = width/2;
    break;
  case 420:
    heightUV = height/2;
    widthUV = width/2;
    break;
  case 422:
    heightUV = height;
    widthUV = width/2;
    break;
  case 444:
    heightUV = height;
    widthUV = width;
    break;
  default:
    fprintf(stderr, "Unknown chroma sampling\n");
    height = width = sampling = 0;
    Y = U = V = Ytemp = NULL;
    comp[0] = comp[1] = comp[2] = NULL;
    return;
  }

  picsizeY = height*width;
  picsizeUV = heightUV*widthUV;
  size_in_byte = picsizeY + picsizeUV*2;

  if(!allocate_mem())
  {
    height = width = sampling = 0;
    Y = U = V = Ytemp = NULL;
    comp[0] = comp[1] = comp[2] = NULL;
    pBuffer = NULL;
  }
}

CIYuv::~CIYuv()
{
  release_mem();
}

bool CIYuv::Resize(int h, int w, int chroma_format)
{
  height = h;
  width = w;
  sampling = chroma_format;

  switch(sampling)
  {
  case 400: // padding chroma components
    heightUV = height/2;
    widthUV  = width/2;
    break;
  case 420:
    heightUV = height/2;
    widthUV  = width/2;
    break;
  case 422:
    heightUV = height;
    widthUV  = width/2;
    break;
  case 444:
    heightUV = height;
    widthUV  = width;
    break;
  default:
    fprintf(stderr, "Unknown chroma sampling\n");
    Y = U = V = Ytemp = NULL;
    return false;
  }

  picsizeY = height*width;
  picsizeUV = heightUV*widthUV;
  size_in_byte = picsizeY + picsizeUV*2;

  return allocate_mem();
}

/*
 * \brief
 *    Allocate mem for the YUV image.
 *    Physical memory is allocated as a 1D array, pointed by pBuffer
 *    However, in order to facilitate to access Y, U and V components in a 2D way,
 *    the following pointers can be refered:
 *    comp[0] (redundant of Y?)
 *    comp[1] (redundant of U?)
 *    comp[2] (redundant of V?)
 * \return
 *    true: succeed
 *    false: fail
 */

bool CIYuv::allocate_mem()
{
  int h, pos;
  BYTE *buf1D;
  BYTE **buf2D;
  BYTE *buf1D_Y;
  BYTE **buf2D_Y;

  release_mem();

  if( (buf1D=(unsigned char *)malloc(size_in_byte*sizeof(BYTE))) == NULL ) return false;

  if( (buf2D=(BYTE **)malloc((height+heightUV*2)*sizeof(BYTE *))) == NULL )
  {
    free(buf1D);
    return false;
  }
  memset(&buf1D[0], 0, picsizeY);
  memset(&buf1D[picsizeY], 128, picsizeUV*2);

  Y = buf2D;
  U = &(buf2D[height]);
  V = &(buf2D[height+heightUV]);

  for(h=pos=0; h<height; h++, pos+=width)
    Y[h] = &(buf1D[pos]);
  for(h=0; h<heightUV; h++, pos+=widthUV)
    U[h] = &(buf1D[pos]);
  for(h=0; h<heightUV; h++, pos+=widthUV)
    V[h] = &(buf1D[pos]);

  comp[0] = Y;
  comp[1] = U;
  comp[2] = V;

  pBuffer = Y[0]; // buf1D;

  buf1D_Y= (unsigned char *)malloc(size_in_byte*sizeof(BYTE));
  buf2D_Y= (BYTE **)malloc((height)*sizeof(BYTE *));
  memset(&buf1D_Y[0], 0, picsizeY);
  Ytemp= buf2D_Y;
  for(h=pos=0; h<height; h++, pos+=width)
    Ytemp[h] = &(buf1D_Y[pos]);

  return true;
}

void CIYuv::release_mem()
{
  if(Y!=NULL)
  {
    if(Y[0]!=NULL) free(Y[0]);
    free(Y);
    free(Ytemp);
    Y = U = V = Ytemp = NULL;
    comp[0] = comp[1] = comp[2] = NULL;
    pBuffer = NULL;
  }
}



/*
 * \brief
 *    Read one frame from file to buffer
 * \return
 *    true: succeed
 *    false: fail
 *
 */
bool CIYuv::ReadOneFrame(FILE *fp, int frameno)
{
  if(Y==NULL || fp==NULL || pBuffer==NULL) return false;

  if (frameno != -1)
  {
    long offs = long(size_in_byte)*long(frameno);
    fseek(fp, offs, SEEK_SET);
  }

  if(fread(Y[0], size_in_byte, 1, fp)!=1)
  {
    fprintf(stderr, "EOF\n");
    return false;
  }

  return true;
}



bool CIYuv::WriteOneFrame(FILE *fp)
{
  if(Y==NULL || fp==NULL || pBuffer==NULL) return false;

  if(fwrite(Y[0], size_in_byte, 1, fp)!=1)
  {
    fprintf(stderr, "EOF\n");
    return false;
  }

  return true;
}



bool CIYuv::WriteOneFrame_inYUV(FILE *fp)
{
  if(Y==NULL || fp==NULL || pBuffer==NULL) return false;

  int x, y;
  uchar *src;
  src = pBuffer;

  switch(sampling)
  {
  case 444:
    for (y=0;y<height;y++) {
      for (x=0;x<width;x++) {
        fputc(src[y*width*3 + x*3 + 0], fp);
      }
    }
    for (y=0;y<height;y++) {
      for (x=0;x<width;x++) {
        fputc(src[y*width*3 + x*3 + 1], fp);
      }
    }
    for (y=0;y<height;y++) {
      for (x=0;x<width;x++) {
        fputc(src[y*width*3 + x*3 + 2], fp);
      }
    }
    break;
  case 420:
    for (y=0;y<height;y++) {
      for (x=0;x<width;x++) {
        fputc(src[y*width*3 + x*3 + 0], fp);
      }
    }
    for (y=0;y<height/2;y++) {
      for (x=0;x<width/2;x++) {
        fputc(src[(y*2)*(width*3) + (x*2)*3 + 1], fp);
      }
    }
    for (y=0;y<height/2;y++) {
      for (x=0;x<width/2;x++) {
        fputc(src[(y*2)*(width*3) + (x*2)*3 + 2], fp);
      }
    }
    break;
  case 400:
    for (y=0;y<height;y++) {
      for (x=0;x<width;x++) {
        fputc(src[y*width + x + 0], fp);
      }
    }
    break;
  }

  return true;
}



bool CIYuv::depthEnhanced( CIYuv *yuvSrc, unsigned int th ) {
  memcpy(Y[0], yuvSrc->Y[0], picsizeY);

  for ( int h= 0; h < height; h++ ) {
    for ( int w= 0; w < width - 7; w++ ) {

      if (
        abs(yuvSrc->Y[h][w] - yuvSrc->Y[h][w+2]) < th
        && abs(yuvSrc->Y[h][w+5] - yuvSrc->Y[h][w+7]) < th
        && abs(yuvSrc->Y[h][w+1] - yuvSrc->Y[h][w+6]) > th // Sep4,'13 5 -> 6
      ) {
        if ( abs(yuvSrc->Y[h][w+2] - yuvSrc->Y[h][w+3]) <= abs(yuvSrc->Y[h][w+3] - yuvSrc->Y[h][w+5]) )
          Y[h][w+3]= yuvSrc->Y[h][w+2];
        else
          Y[h][w+3]= yuvSrc->Y[h][w+5];

        if ( abs(yuvSrc->Y[h][w+2] - yuvSrc->Y[h][w+4]) <= abs(yuvSrc->Y[h][w+4] - yuvSrc->Y[h][w+5]) )
          Y[h][w+4]= yuvSrc->Y[h][w+2];
        else
          Y[h][w+4]= yuvSrc->Y[h][w+5];
      }
    }
  }

  return true;
}



bool CIYuv::setData444_inIYUV(CIYuv *yuvSrc)
{
  int h, dstH, dstW, srcH, srcW;

  if(sampling!=444 || Y==NULL || height!=yuvSrc->getHeight() || width!=yuvSrc->getWidth()) return false;

  memcpy(Y[0], yuvSrc->Y[0], picsizeY);

  switch(yuvSrc->getSampling())
  {
  case 400:
    memset(U[0], 128, picsizeUV*2);
    break;
  case 420:
    for(srcH=dstH=0; dstH<height; dstH+=2, srcH++)
    {
      for(srcW=dstW=0; dstW<width; dstW+=2, srcW++)
      {
        U[dstH][dstW] = U[dstH+1][dstW] = U[dstH][dstW+1] = U[dstH+1][dstW+1] = yuvSrc->U[srcH][srcW];
        V[dstH][dstW] = V[dstH+1][dstW] = V[dstH][dstW+1] = V[dstH+1][dstW+1] = yuvSrc->V[srcH][srcW];
      }
    }
    break;
  case 422:
    for(h=0; h<height; h++)
    {
      for(srcW=dstW=0; dstW<width; dstW+=2, srcW++)
      {
        U[h][dstW] = U[h][dstW+1] = yuvSrc->U[h][srcW];
        V[h][dstW] = V[h][dstW+1] = yuvSrc->V[h][srcW];
      }
    }
    break;
  case 444:
    memcpy(U[0], yuvSrc->U[0], picsizeUV*2);
    break;
  default:
    return false;
  }

  return true;
}



bool CIYuv::setData444_inIYUV420_onlyLEFTSIDE(CIYuv *yuvSrc, bool half) { // Oct25,2012
  int dstH, dstW, srcH, srcW, dstUVH, dstUVW;

  if ( sampling!=444 || Y==NULL ) return false;
  if ( (!half && height!=yuvSrc->getHeight()) || ( half && height*2 !=yuvSrc->getHeight()) ) return false;
  if ( yuvSrc->getSampling() != 420 ) return false;
  // printf("DEBUG:dH:%i,dW:%i,sH:%i,sW:%i\n", height, width, yuvSrc->getHeight(), yuvSrc->getWidth());

  for ( srcH=dstH=dstUVH= 0; dstH < height; dstH++ ) {
    for ( srcW=dstW=dstUVW= 0; dstW < width; dstW++) {
      Y[dstH][dstW]= yuvSrc->Y[dstH][dstW];
      if ( dstW % 2 == 0 && dstH % 2 == 0 ) {
        U[dstUVH][dstUVW] = U[dstUVH+1][dstUVW] = U[dstUVH][dstUVW+1] = U[dstUVH+1][dstUVW+1] = yuvSrc->U[srcH][srcW];
        V[dstUVH][dstUVW] = V[dstUVH+1][dstUVW] = V[dstUVH][dstUVW+1] = V[dstUVH+1][dstUVW+1] = yuvSrc->V[srcH][srcW];
        dstUVW+=2; srcW++;
      }
    }
    if ( dstH % 2 == 0 ) { dstUVH+=2; srcH++; }
  }
  //printf("DEBUG:dH:%i,dW:%i,sH:%i,sW:%i\n", dstUVH, dstUVW, srcH, srcW);

  return true;
}



// July1,2011 NiCT ///////////// Support only sampling == 420.
void CIYuv::setDataFromDepth(IplImage *imgDepth) {
  unsigned char *bufYUV1, *bufYUV2;
  unsigned int h, w;

  for(h=0; h<height; h+=2) {
    bufYUV1 = (unsigned char *)&(imgDepth->imageData[ h   *imgDepth->widthStep]);
    bufYUV2 = (unsigned char *)&(imgDepth->imageData[(h+1)*imgDepth->widthStep]);
    for(w=0; w<width; w+=2) {
      Y[h  ][w  ]= *bufYUV1++;
      Y[h+1][w  ]= *bufYUV2++;
      Y[h  ][w+1]= *bufYUV1++;
      Y[h+1][w+1]= *bufYUV2++;
      U[h/2][w/2]= 128;
      V[h/2][w/2]= 128;
    }
  }

}



void CIYuv::setDataFromImgYUV(IplImage *imgYUV)
{
  int h, w;
  int iu, iv;
  unsigned char *bufYUV1, *bufYUV2;

  switch(sampling)
  {
  case 400:
    //printf("*400");
    for(h=0; h<height; h++)
    {
      bufYUV1 = (unsigned char *)&(imgYUV->imageData[h*imgYUV->widthStep]);
      for(w=0; w<width; w++)
      {
        Y[h][w] = *bufYUV1;  // Y
        bufYUV1 += 3;
      }
    }
    break;
  case 420:
    //printf("*420");
    for(h=0; h<height; h+=2)
    {
      bufYUV1 = (unsigned char *)&(imgYUV->imageData[ h   *imgYUV->widthStep]);
      bufYUV2 = (unsigned char *)&(imgYUV->imageData[(h+1)*imgYUV->widthStep]);
      for(w=0; w<width; w+=2)
      {
        Y[h  ][w  ] = *bufYUV1++;
        iu = *bufYUV1++;
        iv = *bufYUV1++;

        Y[h+1][w  ] = *bufYUV2++;
        iu += *bufYUV2++;
        iv += *bufYUV2++;

        Y[h  ][w+1] = *bufYUV1++;
        iu += *bufYUV1++;
        iv += *bufYUV1++;

        Y[h+1][w+1] = *bufYUV2++;
        iu += *bufYUV2++;
        iv += *bufYUV2++;

        U[h/2][w/2] = CLIP3((iu + 2)/4, 0, 255);
        V[h/2][w/2] = CLIP3((iv + 2)/4, 0, 255);
      }
    }
    break;
  case 422:
    //printf("*422");
    for(h=0; h<height; h++)
    {
      bufYUV1 = (unsigned char *)&(imgYUV->imageData[h*imgYUV->widthStep]);
      for(w=0; w<width; w+=2)
      {
        Y[h][w  ] = *bufYUV1++;
        iu = *bufYUV1++;
        iv = *bufYUV1++;

        Y[h][w+1] = *bufYUV1++;
        iu += *bufYUV1++;
        iv += *bufYUV1++;

        U[h][w/2] = CLIP3((iu + 1)/2, 0, 255);
        V[h][w/2] = CLIP3((iv + 1)/2, 0, 255);
      }
    }
    break;
  case 444:
    //printf("*444");
    for(h=0; h<height; h++)
    {
      bufYUV1 = (unsigned char *)&(imgYUV->imageData[h*imgYUV->widthStep]);
      for(w=0; w<width; w++)
      {
        Y[h][w] = *bufYUV1++;
        U[h][w] = *bufYUV1++;
        V[h][w] = *bufYUV1++;
      }
    }
    break;
  }
}



void CIYuv::setTextureFromImgYUV444( IplImage *imgYUV ) {
  unsigned int h, w;
  unsigned char *bufYUV1;

  for( h=0; h<height; h++ ) {
    bufYUV1= (unsigned char *)&(imgYUV->imageData[h*imgYUV->widthStep]);
    for ( w=0; w<width; w++ ) {
      Y[h][w]= *bufYUV1++;
      U[h][w]= *bufYUV1++;
      V[h][w]= *bufYUV1++;
    }
  }
}



void CIYuv::setDepthFromImgYUV444( IplImage *imgYUV ) {
  unsigned int h, w;
  unsigned char *bufYUV1;

  for( h=0; h<height; h++ ) {
    bufYUV1= (unsigned char *)&(imgYUV->imageData[h*imgYUV->widthStep]);
    for ( w=0; w<width; w++ ) {
      Y[h][w]= *bufYUV1++;
    }
  }
}



bool CIYuv::setUpsampleFilter(unsigned int filter, unsigned int precision)
{
  switch(precision)
  {
  case 1: // INTEGER PEL
    FuncFilter = DummyFilter_2D;
    break;
  case 2: // HALF PEL
    switch(filter)
    {
    case 0: // BI-LINEAR
      FuncFilter = HorizontalLinearFilter_2D_half;
      break;
    case 1: // BI-CUBIC
      FuncFilter = HorizontalCubicFilter_2D_half;
      break;
    case 2: // AVC
      FuncFilter = HorizontalAVCFilter_2D_half;
      break;
    default:
      return false;
      break;
    }
    break;
  case 4: // QUARTER PEL
    switch(filter)
    {
    case 0: // BI-LINEAR
      FuncFilter = HorizontalLinearFilter_2D_qpel;
      break;
    case 1: // BI-CUBIC
      FuncFilter = HorizontalCubicFilter_2D_qpel;
      break;
    case 2: // AVC
      FuncFilter = HorizontalAVCFilter_2D_qpel;
      break;
    default:
      return false;
      break;
    }
    break;
  default:
    return false;
    break;
  }

  return true;
}



bool CIYuv::upsampling(CIYuv *src, int padding_size)
{
  // need to check buffer size

  FuncFilter(src->Y, Y, src->getWidth(), src->getHeight(), padding_size);
  FuncFilter(src->U, U, src->getWidthUV(), src->getHeightUV(), padding_size);
  FuncFilter(src->V, V, src->getWidthUV(), src->getHeightUV(), padding_size);

  return true;
}



bool CIYuv::BQF_V(CIYuv *src)
{
  VerticalCubicFilter_2D_half_V(src->Y, Y, src->getWidth(), src->getHeight());
  VerticalCubicFilter_2D_half_V(src->U, U, src->getWidthUV(), src->getHeightUV());
  VerticalCubicFilter_2D_half_V(src->V, V, src->getWidthUV(), src->getHeightUV());

  return true;
}



bool CIYuv::BQF_V3(CIYuv *src, unsigned int side)
{
  VerticalCubicFilter_2D_quart_V(src->Y, Y, src->getWidth(), src->getHeight(), side);
  VerticalCubicFilter_2D_quart_V(src->U, U, src->getWidthUV(), src->getHeightUV(), side);
  VerticalCubicFilter_2D_quart_V(src->V, V, src->getWidthUV(), src->getHeightUV(), side);

  return true;
}



bool CIYuv::BQF_D(CIYuv *src)
{
  VerticalCubicFilter_2D_half_D(src->Y, Y, src->getWidth(), src->getHeight());
  VerticalCubicFilter_2D_half_D(src->U, U, src->getWidthUV(), src->getHeightUV());
  VerticalCubicFilter_2D_half_D(src->V, V, src->getWidthUV(), src->getHeightUV());

  return true;
}



bool CIYuv::BQF_D3(CIYuv *src, unsigned int side)
{
  VerticalCubicFilter_2D_quart_D(src->Y, Y, src->getWidth(), src->getHeight(), side);
  VerticalCubicFilter_2D_quart_D(src->U, U, src->getWidthUV(), src->getHeightUV(), side);
  VerticalCubicFilter_2D_quart_D(src->V, V, src->getWidthUV(), src->getHeightUV(), side);

  return true;
}



void CIYuv::BQF_V_Expand(CIYuv *src) {
  VerticalCubicFilter_H(src->Y, Y, src->getWidth(), src->getHeight());
  VerticalCubicFilter_H(src->U, U, src->getWidthUV(), src->getHeightUV());
  VerticalCubicFilter_H(src->V, V, src->getWidthUV(), src->getHeightUV());
}



void CIYuv::BQF_V_Single(CIYuv *src) {
  VerticalCubicFilter_Single(src->Y, Y, src->getWidth(), src->getHeight());
  VerticalCubicFilter_Single(src->U, U, src->getWidthUV(), src->getHeightUV());
  VerticalCubicFilter_Single(src->V, V, src->getWidthUV(), src->getHeightUV());
}



void CIYuv::BQF_V_Split(CIYuv *src, unsigned int side) {
  VerticalCubicFilter_Split(src->Y, Y, src->getWidth(), src->getHeight(), side);
  VerticalCubicFilter_Split(src->U, U, src->getWidthUV(), src->getHeightUV(), side);
  VerticalCubicFilter_Split(src->V, V, src->getWidthUV(), src->getHeightUV(), side);
}



void CIYuv::BQF_H(CIYuv *src) {
  HorizontalCubicFilter(src->Y, Y, src->getWidth(), src->getHeight());
  HorizontalCubicFilter(src->U, U, src->getWidthUV(), src->getHeightUV());
  HorizontalCubicFilter(src->V, V, src->getWidthUV(), src->getHeightUV());
}



void CIYuv::BQF_Single(CIYuv *src) {
  VerticalCubicFilter_2DSingle(src->Y, Y, src->getWidth(), src->getHeight());
  VerticalCubicFilter_2DSingle(src->U, U, src->getWidthUV(), src->getHeightUV());
  VerticalCubicFilter_2DSingle(src->V, V, src->getWidthUV(), src->getHeightUV());
}



void CIYuv::BQF_StackSplit(CIYuv *src, unsigned int side) {
  VerticalCubicFilter_2DStackSplit(src->Y, Y, src->getWidth(), src->getHeight(), side);
  VerticalCubicFilter_2DStackSplit(src->U, U, src->getWidthUV(), src->getHeightUV(), side);
  VerticalCubicFilter_2DStackSplit(src->V, V, src->getWidthUV(), src->getHeightUV(), side);
}



void CIYuv::looseEdgeBindResiduals( CIYuv *map, unsigned int m_mode ) {
  looseEdgeBindData( Y, map->Y, this->getWidth(), this->getHeight() );
  looseEdgeBindData( U, map->Y, this->getWidthUV(), this->getHeightUV() );
  looseEdgeBindData( V, map->Y, this->getWidthUV(), this->getHeightUV() );
}



bool CIYuv::LPF_V(CIYuv *src, bool full) {
  VerticalLowpassFilter3tap(src->Y, Y, src->getWidth(), src->getHeight());
  if (!full) return true;
  VerticalLowpassFilter3tap(src->U, U, src->getWidthUV(), src->getHeightUV());
  VerticalLowpassFilter3tap(src->V, V, src->getWidthUV(), src->getHeightUV());

  return true;
}



bool CIYuv::LPF_H(CIYuv *src, bool full) {
  HorizontalLowpassFilter3tap(src->Y, Y, src->getWidth(), src->getHeight());
  if (!full) return true;
  HorizontalLowpassFilter3tap(src->U, U, src->getWidthUV(), src->getHeightUV());
  HorizontalLowpassFilter3tap(src->V, V, src->getWidthUV(), src->getHeightUV());

  return true;
}



void CIYuv::LPF_V_DExpand(CIYuv *src, unsigned int th, bool full) {
  VerticalLinearFilter_E_3tap(src->Y, Y, src->getWidth(), src->getHeight(), th);
  if (!full) return;
  VerticalLinearFilter_E_3tap(src->U, U, src->getWidthUV(), src->getHeightUV(), th);
  VerticalLinearFilter_E_3tap(src->V, V, src->getWidthUV(), src->getHeightUV(), th);
}



void CIYuv::LPF_V_DSingle(CIYuv *src, unsigned int th, bool full) {
  VerticalLinearFilter_Single_3tap(src->Y, Y, src->getWidth(), src->getHeight(), th);
  if (!full) return;
  VerticalLinearFilter_Single_3tap(src->U, U, src->getWidthUV(), src->getHeightUV(), th);
  VerticalLinearFilter_Single_3tap(src->V, V, src->getWidthUV(), src->getHeightUV(), th);
}



void CIYuv::LPF_V_DSplit(CIYuv *src, unsigned int side, unsigned int th, bool full) {
  // DeblockF(src->Y, src->Ytemp, src->getWidth(), src->getHeight(), side, WIDE );
  // VerticalLinearFilter_Split_3tap(src->Ytemp, Y, src->getWidth(), src->getHeight(), side, th);
  VerticalLinearFilter_Split_3tap(src->Y, Y, src->getWidth(), src->getHeight(), side, th);
  if (!full) return;
  VerticalLinearFilter_Split_3tap(src->U, U, src->getWidthUV(), src->getHeightUV(), side, th);
  VerticalLinearFilter_Split_3tap(src->V, V, src->getWidthUV(), src->getHeightUV(), side, th);
}



void CIYuv::LPF_H_DSmooth(CIYuv *src, unsigned int th, bool full) {
  HorizontalLinearFilter_S_3tap(src->Y, Y, src->getWidth(), src->getHeight(), th);
  if (!full) return;
  HorizontalLinearFilter_S_3tap(src->U, U, src->getWidthUV(), src->getHeightUV(), th);
  HorizontalLinearFilter_S_3tap(src->V, V, src->getWidthUV(), src->getHeightUV(), th);
}



void CIYuv::LPF_V_RExpand(CIYuv *src, unsigned int th) {
  VerticalLinearFilter_E_2tap(src->Y, Y, src->getWidth(), src->getHeight(), th);
  VerticalLinearFilter_E_2tap(src->U, U, src->getWidthUV(), src->getHeightUV(), th);
  VerticalLinearFilter_E_2tap(src->V, V, src->getWidthUV(), src->getHeightUV(), th);
}



void CIYuv::LPF_V_RSingle(CIYuv *src, unsigned int th) {
  VerticalLinearFilter_Single_2tap(src->Y, Y, src->getWidth(), src->getHeight(), th);
  VerticalLinearFilter_Single_2tap(src->U, U, src->getWidthUV(), src->getHeightUV(), th);
  VerticalLinearFilter_Single_2tap(src->V, V, src->getWidthUV(), src->getHeightUV(), th);
}



void CIYuv::LPF_V_RSplit(CIYuv *src, unsigned int side, unsigned int th) {
  VerticalLinearFilter_Split_2tap(src->Y, Y, src->getWidth(), src->getHeight(), side, th);
  VerticalLinearFilter_Split_2tap(src->U, U, src->getWidthUV(), src->getHeightUV(), side, th);
  VerticalLinearFilter_Split_2tap(src->V, V, src->getWidthUV(), src->getHeightUV(), side, th);
}



void CIYuv::LPF_H_RSmooth(CIYuv *src, unsigned int th) {
  HorizontalLinearFilter_S_2tap(src->Y, Y, src->getWidth(), src->getHeight(), th);
  HorizontalLinearFilter_S_2tap(src->U, U, src->getWidthUV(), src->getHeightUV(), th);
  HorizontalLinearFilter_S_2tap(src->V, V, src->getWidthUV(), src->getHeightUV(), th);
}
